// Declare global constants
// var gv_degToRad = Math.PI / 180.0;
var gv_deltaAccel = 0.01;
var gv_deltaAngle = 0.04;

var gv_requestAnimationFrame;
var gv_isGameInProgress = false;

var gv_canvas = document.getElementById("canvas");
var gv_context = gv_canvas.getContext("2d");

var gv_images = {};
var gv_numImagesLoaded = 0;
var gv_numImagesRequested = 1;

var gv_car = null;

/**
 * Log information to the screen.
 *
 * @param text The text to display in the log area.
 */
function log(text) {
    var logArea = document.getElementById("log");
    logArea.innerHTML = text;
}

/**
 * Process key down events.
 * This is how the car's speed is adjusted, how the car is turned, and how the game is ended.
 *
 * @param event The key down event.
 */
function onKeyDown(event) {
    if (event.keyCode === 38) {  // up arrow
        // increase acceleration
        gv_car.isAccelerating = true;
    } else if (event.keyCode === 40) {  // down arrow
        // decrease deceleration
        gv_car.isDecelerating = true;
    } else if (event.keyCode === 37) {  // left arrow
        // turn left
        gv_car.turningLeft = true;
    } else if (event.keyCode === 39) {  // right arrow
        // turn right
        gv_car.turningRight = true;
    } else if (event.keyCode === 32) {  // space key
        gv_car.hasFriction = true;
    } else if (event.keyCode === 27) {  // escape key
        // End the game
        window.cancelAnimationFrame(gv_requestAnimationFrame);
        log("<span class='gameover'>Game Over<\/span>");
        gv_isGameInProgress = false;
    }
}

/**
 * Process key up events.
 * This is how the car's speed is adjusted, how the car is turned, and how the game is ended.
 *
 * @param event The key up event.
 */
function onKeyUp(event) {
    if (event.keyCode === 38) {  // up arrow
        // increase speed
        gv_car.isAccelerating = false;
    } else if (event.keyCode === 40) {  // down arrow
        // decrease speed
        gv_car.isDecelerating = false;
    } else if (event.keyCode === 37) {  // left arrow
        // turn left
        gv_car.turningLeft = false;
    } else if (event.keyCode === 39) {  // right arrow
        // turn right
        gv_car.turningRight = false;
    } else if (event.keyCode === 32) {  // space key
        gv_car.hasFriction = false;
    } else if (event.keyCode === 27) {  // escape key
        // End the game
    }
}

/**
 * Define a Car object.
 *
 * @param image The image used to represent the car on screen.
 * @constructor Defines a Car object.
 */
function Car(image) {
    this.image = image;
    this.x = image.width / 2;
    this.y = image.height / 2;
    this.speed = 0;
    this.angle = 0;
    this.friction = 0.95;
    this.turningLeft = false;
    this.turningRight = false;
    this.isAccelerating = false;
    this.isDecelerating = false;
    this.hasFriction = false;

    /**
     * Draw the car at its current location and rotation.
     */
    this.draw = function() {
        gv_context.save();
        gv_context.translate(this.x, this.y);
        gv_context.rotate(this.angle);
        gv_context.drawImage(this.image, -this.image.width / 2, -this.image.height / 2);
        gv_context.restore();
    };

    /**
     * Update the car's speed, rotation, and position.
     */
    this.update = function() {
        if (this.isAccelerating) {
            this.speed += gv_deltaAccel;
        } else if (this.isDecelerating) {
            this.speed -= gv_deltaAccel;
        }

        if (this.hasFriction) {
            this.speed *= this.friction;
        }

        if ((this.speed > -0.005) && (this.speed < 0.005)) {
            this.speed = 0.0;  // Ensure braking to zero.
        }

        if ((this.speed >= 0.01) || (this.speed <= -0.01)) {
            if (this.turningLeft) {
                this.angle -= gv_deltaAngle;
            } else if (this.turningRight) {
                this.angle += gv_deltaAngle;
            }
        }

        this.x += this.speed * Math.cos(this.angle);
        this.y += this.speed * Math.sin(this.angle);

        if (this.x > gv_canvas.width) {
            this.x -= gv_canvas.width;
        } else if (this.x < 0) {
            this.x += gv_canvas.width;
        }
        if (this.y > gv_canvas.height) {
            this.y -= gv_canvas.height;
        } else if (this.y < 0) {
            this.y += gv_canvas.height;
        }
    };
}

/**
 * Load an image given the name of the PNG image file.
 * @param imageName The name of the image file without its file suffix (assumes a ".png" file suffix).
 */
function loadImage(imageName) {
	var image = new Image();
    gv_images[imageName] = image;

	image.onload = onImageLoaded;
	image.src = "images/" + imageName + ".png";
}

/**
 * This function is called once every update of the animation frame.
 * It draws the background, updates the car position, draws the car,
 * and logs the car's speed and angle of rotation.
 */
function updateGame() {
    if (! gv_isGameInProgress) {
        return;
    }

    // Draw the background
    gv_context.fillStyle = "#ECECEC";
    gv_context.fillRect(0, 0, gv_canvas.width, gv_canvas.height);
    gv_context.strokeStyle = "#7F7F7F";
    gv_context.strokeRect(0, 0, gv_canvas.width, gv_canvas.height);

    gv_car.update();
    gv_car.draw();

    log("Current speed = " + gv_car.speed.toFixed(2) + " at " + utils.toDegrees(gv_car.angle).toFixed(2) + " degrees.");  // Convert speed to units
}

/**
 * Initialize the game.
 */
function initGame() {
    gv_car = new Car(gv_images.car1);
    gv_isGameInProgress = true;

    (function drawFrame() {
        gv_requestAnimationFrame = window.requestAnimationFrame(drawFrame, gv_canvas);
        updateGame();
    }());

    window.addEventListener("keydown", onKeyDown, false);
    window.addEventListener("keyup", onKeyUp, false);
}

/**
 * This function is called once for each image loaded.
 */
function onImageLoaded() {
    gv_numImagesLoaded++;

    // start game only after we know all images have been loaded.
    if (gv_numImagesLoaded == gv_numImagesRequested) {
        initGame();
    }
}

loadImage("car1");  // Load the image of the car.  This starts the game.
